import 'dart:io';
import 'dart:math';

import 'package:collection/collection.dart';
import 'package:perkataan/src/math_utils.dart';

/// The [Words] class keeps a cache of words with methods to analyze those
/// words and produce appropriate results.
class Words {
  // The string are wrapped within a [_Word] class and we map the word length
  // to a list of entries.
  final Map<int, List<_Word>>_wordsByLength = Map();

  /// A default word list is provided within data folder
  Words.all() {
    _load(new File("data/allWords.txt").readAsLinesSync());
  }

  /// Load a specific list of words.
  Words.from(List<String> words) {
    _load(words);
  }

  void _load(Iterable<String> words) {
    words.forEach((w) => _add(w));
  }

  // The word is trimmed to remove white space and if the resulting string is
  // not empty it is wrapped and added to the word cache.
  void _add(String word) {
    if (word != null) {
      String toAdd = word.trim();

      if (toAdd.length > 0) {
        _wordsByLength.putIfAbsent(word.length, () => List()).add(_Word(word));
      }
    }
  }

  /// Total number of words we have stored.
  int numOfWords() {
    int res = 0;

    for (var words in _wordsByLength.values) {
      res += words.length;
    }

    return res;
  }

  /// Returns the number of words with a specific length.
  int size(int length) {
    var list = _wordsByLength[length];

    if (list != null) {
      return list.length;
    } else {
      return 0;
    }
  }

  /// Anagram generator returning all words that contain the provided letters
  /// The following code
  ///    words.anagrams('rode')
  /// would produce the following list
  /// ['doer', 'oder', 'redo', 'rode']
  Iterable<String> anagrams(String letters) sync* {
    var words = _wordsByLength[letters.length];

    if (words != null) {
      for (var word in words) {
        if (word.isAnagram(letters)) {
          yield word.word;
        }
      }
    }
  }

  /// SubAnagram generator returning words that contain the provided letters or
  /// a subset with a minimum size of [minSize].
  Iterable<String> subAnagrams(int minSize, String letters) sync* {
    if (minSize > letters.length) {
      throw new ArgumentError.value(minSize, "minSize", "Must be >= letters length");
    }

    for (int index = minSize; index <= letters.length; index++ ) {
      for (var word in _wordsByLength[index]) {
        if (word.isSubAnagram(letters))
          yield word.word;
      }
    }
  }

  /// Generator of random words of length within a given range from [minSize] inclusive to [maxSize] inclusive.
  Iterable<String> randomWords(int minSize, int maxSize) sync* {
    var rnd = Random();
    var rangeRnd = randomIntsInclusiveRange(minSize, maxSize);

    for (var next in rangeRnd) {
      // Pick a random word length.
      var words = _wordsByLength[next];

      if (words != null) {
        yield words[rnd.nextInt(words.length)].word;
      }
    }
  }

  /// Generator to produce words that have given letters in same position.
  /// Typically this is a crossword solver.
  /// The [needed] input maps pos => letter code.
  Iterable<String> wordsMatching(int length, Map<int, int> needed) sync* {
    for (var word in _wordsByLength[length]) {
      if (word.hasLettersAt(needed)) {
        yield word.word;
      }
    }
  }
}

class _Word {
  List<int> encoded;

  _Word(String word) {
    encoded = word.toLowerCase().codeUnits;
  }

  String get word => String.fromCharCodes(encoded);

  bool isAnagram(String toMatch) {
    return UnorderedIterableEquality().equals(encoded, toMatch.toLowerCase().codeUnits);
  }

  bool isSubAnagram(String toMatch) {
    Map<int, int> wordMakeup = Map();

    for (var letter in toMatch.toLowerCase().codeUnits) {
      int count = wordMakeup[letter];

      if (count == null) count = 0;
      wordMakeup[letter] = count + 1;
    }

    for (var letter in encoded) {
      int count = wordMakeup[letter];
      if (count == null || count == 0) return false;
      wordMakeup[letter] = count - 1;
    }

    return true;
  }

  bool containsAll(List<int> letterCodes) {
    // todo short-ciruit when more letterCodes than letters to check.
    for (var letter in encoded) {
      if (!letterCodes.remove(letter)) {
        return false;
      }
    }

    return true;
  }

  /**
   * entry is a map of pos => letter
   */
  bool hasLettersAt(Map<int, int> needed) {
    return needed.entries.every((entry) => (entry.key >= 0 && entry.key < encoded.length && encoded[entry.key] == entry.value));
  }
}